import mosaik_api_v3
from midas.util.dict_util import bool_from_dict
from midas.util.logging import set_and_init_logger
from mosaik.exceptions import SimulationError
from mosaik_api_v3.types import (
    CreateResult,
    EntityId,
    Meta,
    ModelName,
    OutputData,
    OutputRequest,
    SimId,
)
from typing_extensions import override

from . import LOG
from .csv_model import CSVModel
from .meta import META


class MidasCSVStore(mosaik_api_v3.Simulator):
    """Simulator to store simulation results in a csv file."""

    def __init__(self) -> None:
        super().__init__(META)

        self.sid: SimId | None = None
        self.eid: EntityId = "Database-0"
        self.database: CSVModel | None = None

        self.filename: str | None = None
        self.step_size: int = 0
        self.current_size: int = 0
        self.saved_rows: int = 0
        self.finalized: bool = False
        self.keep_old_files: bool = True

    @override
    def init(
        self, sid: SimId, time_resolution: float = 1.0, **sim_params
    ) -> Meta:
        self.sid = sid
        self.step_size = sim_params.get("step_size", 900)

        return self.meta

    @override
    def create(
        self, num: int, model: ModelName, **model_params
    ) -> list[CreateResult]:
        if num > 1 or self.database is not None:
            errmsg = (
                "You should really not try to instantiate more than one "
                "database. If your need another database, create a new "
                "simulator as well."
            )
            raise ValueError(errmsg)

        self.database = CSVModel(
            model_params.get("filename", ""),
            path=model_params.get("path", None),
            unique_filename=bool_from_dict(
                model_params, "unique_filename", False
            ),
            keep_old_files=bool_from_dict(
                model_params, "keep_old_files", False
            ),
        )

        return [{"eid": self.eid, "type": model}]

    @override
    def step(self, time, inputs, max_advance=0):
        if self.database is None:
            msg = "Database is unexpectedly None. Can not proceed any further"
            raise SimulationError(msg)

        data = inputs.get(self.eid, {})

        if not data:
            LOG.info(
                "Did not receive any inputs. "
                "Did you connect anything to the store?"
            )

        for attr, src_ids in data.items():
            for src_id, val in src_ids.items():
                sid, eid = src_id.split(".")
                self.database.to_memory(sid, eid, attr, val)

        self.database.step()

        return time + self.step_size

    @override
    def get_data(self, outputs: OutputRequest) -> OutputData:
        return {}

    @override
    def finalize(self):
        LOG.info("Finalizing database.")
        if self.database is not None:
            self.database.finalize()


if __name__ == "__main__":
    set_and_init_logger(0, "store-logfile", "midas-store.log", replace=True)

    mosaik_api_v3.start_simulation(MidasCSVStore())



import csv
import json
import logging
import multiprocessing as mp
import os
import queue
import traceback
from multiprocessing import Queue
from multiprocessing.context import SpawnProcess
from uuid import uuid4

import numpy as np
from midas.util.dict_util import convert_val
from midas.util.runtime_config import RuntimeConfig
from mosaik.exceptions import SimulationError

LOG = logging.getLogger(__name__)


class CSVModel:
    def __init__(
        self,
        filename: str,
        *,
        path: str | None = None,
        unique_filename: bool = False,
        keep_old_files: bool = False,
    ) -> None:
        if path is None:
            path = RuntimeConfig().paths["output_path"]
            if path is None:
                path = ""
        os.makedirs(path, exist_ok=True)

        self.filename = os.path.abspath(os.path.join(path, filename))

        if self.filename and unique_filename:
            fp, suf = self.filename.rsplit(".", 1)
            self.filename = f"{fp}-{str(uuid4())}.{suf}"
        elif not self.filename:
            self.filename = f"midas-store-results-{str(uuid4())}.csv"

        if keep_old_files:
            LOG.debug(
                "Keep_old_files is set to True. Attempting to find a unique "
                "filename for the database."
            )
            incr = 2
            new_filename = self.filename
            while os.path.exists(new_filename):
                fp, suf = self.filename.rsplit(".", 1)
                new_filename = f"{fp}_{incr:03d}.{suf}"
                incr += 1
            self.filename = new_filename
        elif os.path.exists(self.filename):
            os.rename(self.filename, f"{self.filename}.old")

        LOG.info("Saving results to database at '%s'.", self.filename)
        self._columns: list[str] = []
        self._data: dict[str, str | bool | int | float | None] = {}
        self._ctr: int = 0

        self._ctx = mp.get_context("spawn")
        self._io_proc: SpawnProcess | None = None
        self._queue: Queue = self._ctx.Queue()
        self._result: Queue = self._ctx.Queue()

    def to_memory(self, sid, eid, attr, vals):
        key = build_column_key(sid, eid, attr)
        self._data[key] = serialize(vals)

    def step(self):
        if self._io_proc is None:
            LOG.debug("Starting file writer process ...")
            self._columns = list(self._data)
            self._io_proc = self._ctx.Process(
                target=run_writer,
                args=(self.filename, self._columns, self._queue, self._result),
            )
            self._io_proc.start()

        if not self._result.empty():
            msg = "Writer process terminated early. Can't continue from here."
            raise SimulationError(msg)

        self._queue.put(self._data)

    def finalize(self):
        LOG.info("Shutting down the writer process ...")
        if self._io_proc is None:
            LOG.info("Writer is already None (likely was never initialized).")
        else:
            try:
                self._queue.put(-1)
            except ValueError:
                LOG.debug("Queue was already closed.")

            LOG.debug("Waiting for writer to finish ...")
            self._io_proc.join()
            LOG.debug("Writer finished. Check if it left a message ...")

            try:
                msg = self._result.get(timeout=1)
            except queue.Empty:
                LOG.info("Writer finished without message.")
            else:
                LOG.info("Writer finished with message: %s", msg)

        self._result.close()
        self._queue.close()


def run_writer(filename, fields, lines: Queue, result: Queue, timeout=300):
    res_msg = "Finished successfully."
    to_ctr = timeout
    try:
        with open(filename, "w", newline="") as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=fields)
            writer.writeheader()

            while True:
                try:
                    item = lines.get(block=True, timeout=1)
                except queue.Empty:
                    to_ctr -= 1
                    if to_ctr <= 0:
                        res_msg = (
                            f"Writer received no item in the last {timeout} "
                            "seconds."
                        )
                        break
                    continue
                except ValueError:
                    res_msg = "Queue was closed. Terminating!"
                    break

                if isinstance(item, int):
                    LOG.info("Received -1. Terminating!")
                    break

                writer.writerow(item)
                to_ctr = timeout

    except Exception:
        res_msg = f"Error writing csv: {traceback.format_exc()}"
    except KeyboardInterrupt:
        res_msg = "Interrupted by user!"
    try:
        result.put(res_msg)
    except ValueError:
        LOG.info("Result queue was already closed.")


def build_column_key(sid, eid, attr) -> str:
    return f"{sid}.{eid}.{attr}"


def serialize(val):
    new_val = convert_val(val)

    if new_val == "MISSING_VALUE":
        if isinstance(val, (list, dict, np.ndarray)):
            return json.dumps(val)

    return new_val



META = {
    "type": "time-based",
    "models": {
        "DatabaseCSV": {
            "public": True,
            "any_inputs": True,
            "params": [
                "filename",
                "verbose",
                "path",
                "keep_old_files",
                "unique_filename",
            ],
            "attrs": [],
        }
    },
}



import os
import unittest

from midas_store import simulator


class TestMidasCSV(unittest.TestCase):
    def setUp(self):
        self.inputs1 = {
            "Database-0": {
                "p_mw": {
                    "DummySim-0.DummyHousehold-0": 0.1,
                    "DummySim-0.DummyHousehold-1": 0.2,
                    "DummySim-1.DummyPV-0": 0.3,
                },
                "q_mvar": {
                    "DummySim-0.DummyHousehold-0": 0.01,
                    "DummySim-0.DummyHousehold-1": 0.02,
                    "DummySim-1.DummyPV-0": 0.03,
                },
                "t_air": {"DummyWeather-0.WeatherCurrent-0": 15.0},
                "schedule": {"DummySim-2.DummyCHP-0": [10, 12, 10]},
            }
        }
        self.inputs2 = {
            "Database-0": {
                "p_mw": {
                    "DummySim-0.DummyHousehold-0": 0.02,
                    "DummySim-0.DummyHousehold-1": 0.02,
                    "DummySim-1.DummyPV-0": 0.03,
                    "DummySim-2.DummyCHP-0": 0.5,
                },
                "q_mvar": {
                    "DummySim-0.DummyHousehold-0": 0.01,
                    "DummySim-0.DummyHousehold-1": 0.015,
                    "DummySim-1.DummyPV-0": 0.01,
                },
                "t_air": {"DummyWeather-0.WeatherCurrent-0": 15.0},
                "wind": {"DummyWeather-1.WeatherForecast-0": 20},
            }
        }

    def test_setup(self):
        """Test store creation and ensure to allow only one instance."""
        dbsim = simulator.MidasCSVStore()

        dbsim.init("MidasCSVStore", step_size=900)

        # Only one instance allowed
        with self.assertRaises(ValueError):
            dbsim.create(2, "DatabaseCSV", filename="there.csv")

        dbsim.create(1, "DatabaseCSV", filename="here.csv")

        self.assertIsNotNone(dbsim.database)
        if dbsim.database is not None:
            self.assertIn("here.csv", dbsim.database.filename)
        self.assertEqual(900, dbsim.step_size)

        # Only one instance allowed
        with self.assertRaises(ValueError):
            dbsim.create(1, "DatabaseCSV", filename="there.csv")

    def test_step(self):
        dbsim = simulator.MidasCSVStore()
        dbsim.init("MidasCSVStore", step_size=900)
        dbsim.create(1, "DatabaseCSV", filename="here.csv")
        dbsim.step(0, self.inputs1)

        with self.assertLogs("midas_store.csv_model", level="INFO") as cm:
            dbsim.step(900, self.inputs2)
            dbsim.finalize()
        self.assertIn("dict contains fields not in fieldnames", cm.output[-1])

    def test_step_no_inputs(self):
        dbfile = "not-there.csv"
        dbsim = simulator.MidasCSVStore()
        dbsim.init("MidasCSVStore", step_size=900)
        dbsim.create(1, "DatabaseCSV", filename=dbfile)
        dbsim.step(0, {})
        dbsim.finalize()
        self.assertFalse(os.path.exists(dbfile))

    def test_huge_dataset(self):
        """Test if a large dataset can be stored. Takes very long
        and should not be necessary most of the time.
        """
        dbsim = simulator.MidasCSVStore()
        dbsim.init("MidasCSVStore", step_size=900)
        dbsim.create(1, "DatabaseCSV", filename="this_is_huge.csv")

        for idx in range(5 * 365 * 24 * 4):
            dbsim.step(idx * 900, self.inputs1)

            if idx % 96 == 0:
                print(idx / 96, end="\r")

        print()
        dbsim.finalize()


if __name__ == "__main__":
    unittest.main()



"""MIDAS scenario upgrade module.

This module adds a mosaikhdf database to the scenario.

"""

import logging
import os

from midas.scenario.upgrade_module import UpgradeModule
from midas.util.dict_util import set_default_bool

LOG = logging.getLogger(__name__)


class DatabaseModule(UpgradeModule):
    def __init__(self):
        super().__init__(
            module_name="store",
            default_scope_name="database",
            default_sim_config_name="MidasStore",
            default_import_str="midas_store.simulator:MidasCSVStore",
            default_cmd_str=("%(python)s -m midas_store.simulator %(addr)s"),
            log=LOG,
        )
        self.default_filename = "midas_store.csv"
        self._filename = None
        self._path = None
        self._unique_filename = False
        self._keep_old_files = False

    def check_module_params(self, module_params):
        """Check module params for this upgrade."""

        module_params.setdefault(self.default_scope_name, dict())
        module_params.setdefault("filename", self.default_filename)
        module_params.setdefault("path", self.scenario.base.output_path)
        set_default_bool(module_params, "unique_filename", False)
        set_default_bool(module_params, "keep_old_files", False)

    def check_sim_params(self, module_params):
        self._filename = module_params["filename"]
        self._path = module_params["path"]
        self._unique_filename = module_params["unique_filename"]
        self._keep_old_files = module_params["keep_old_files"]

    def start_models(self):
        mod_key = "database"
        params = {
            "filename": self._filename,
            "path": self._path,
            "unique_filename": self._unique_filename,
            "keep_old_files": self._keep_old_files,
        }

        self.start_model(mod_key, "DatabaseCSV", params)

    def connect(self):
        pass

    def connect_to_db(self):
        pass
